import 'package:flutter/material.dart';
import 'package:hypi_tink_api/hypi_tink_api.dart';

/// Added to functions that should be used as the handler for an event.
/// If two handlers are found in the same class, they are invoked on the same instance of the class
/// IFF the actions are triggered on the same widget or a child widget that has had the instance passed to it
class TinkActionFunction {
  /// A short friendly name to display in the UI e.g. a function called findAllMyAccountUsers could have the label "Find all users"
  final String label;

  /// A longer description of the action that is shown to the user in help dialogs
  final String description;

  ///A list of the  fields that the object returned by the function can product. The returned object must implement [Mappable]
  final List<TinkParam>? outputFieldNames;

  /// A unique ID for the action. If not set, this will default to package.Class.functionName
  final String? id;

  const TinkActionFunction(
    this.label,
    this.description, {
    this.outputFieldNames,
    this.id,
  });

  @override
  bool operator ==(Object other) =>
      identical(this, other) ||
      other is TinkActionFunction &&
          runtimeType == other.runtimeType &&
          id == other.id;

  @override
  int get hashCode => id.hashCode;
}

/*
class TinkOutputStates {
  final List<TinkParam> outputs;

  const TinkOutputStates(this.outputs);
}
*/

mixin TinkApplicationState<
    I extends TinkWidgetInputBinding,
    O extends TinkWidgetOutputBinding,
    W extends TinkWorkflowBinding,
    V extends TinkVariable> {
  ///Get all device variable references. Note that the value of the variables are not included
  Map<String, V> get allDeviceVariables;

  ///Gets the reference to the a device variable of the given name
  V? getDeviceVariableRef(String name);

  ///Get all app variable references. Note that the value of the variables are not included
  Map<String, V> get allAppVariables;

  ///Gets the reference to the a device variable of the given name
  V? getAppVariableRef(String name);

  ///Get all app variable references. Note that the value of the variables are not included
  Map<String, V>? allScreenVariables(TinkScreenId screenPath);

  ///Gets the reference to the a device variable of the given name
  V? getScreenVariableRef(
    TinkScreenId screenId,
    String name,
  );
}

class ActionEnv<A extends TinkApplicationState, S extends ScreenModel> {
  ///A map where arbitrary data can be stored by an action
  ///For example, if a custom config widget is used, put custom state in here
  ///All values in this MUST be serialisable (int, double, string, map, list)
  final Map<TinkWorkflowBinding, Map<String, dynamic>> data = {};

  ///Can be used to resolve variables or get references to them
  A state;

  Map<String, S> get screens => _screens();
  final Map<String, S> Function() _screens;

  Map<String, S> get components => _components();
  final Map<String, S> Function() _components;

  ActionEnv(this.state, this._screens, this._components);
}

mixin TinkActionLike<A extends TinkApplicationState, W extends WorkflowModel,
    S extends ScreenModel> implements Displayable {
  late TinkActionLike<A, W, S> delegate;

  List<TinkParam>? get inputs;

  List<TinkParam> get outputs;

  void next(ActionChain chain);

  Widget config(
    ActionEnv<A, S> env,
    W workflow,
    TinkScreenId screenId,
    TinkWorkflowBinding binding,
  );

  @override
  bool operator ==(Object other) =>
      identical(this, other) ||
      other is TinkActionLike &&
          runtimeType == other.runtimeType &&
          id == other.id;

  @override
  int get hashCode => id.hashCode;

  @override
  String toString() {
    final str = StringBuffer();
    inputs?.forEach((i) => str.write(i.toString() + ','));
    str.write(' => ');
    for (final i in outputs) {
      str.write(i.toString() + ',');
    }
    return str.toString();
  }
}

mixin ActionChain<A extends TinkApplicationState, S extends ScreenModel> {
  ActionEnv<A, S> get env;

  BuildContext get context;

  TinkWorkflowBinding get binding;

  ///This allows an action to emit all its output states. It cannot emit any state it hasn't declared, doing so will potentially lead to unexpected behaviour so this will ignore any unknown states emitted
  ///Try to keep names simple and semantic - they're visible to the user and should imply their content/use e.g. user_id is better than uid
  ///Emitting a state with the same name as a previous state that has the same type will overwrite the previously emitted value
  ///[action] is the action that wants to emit the state. It determines which states are can be emitted. An action can only emit variables in its input and output list
  void emitState(TinkActionLike action, TinkVariableData state);

  ///Gets the current value of the given state
  ///[action] is the action that wants to get the state. It determines which states are returned. An action can only get variables in its input and output list
  TinkVariableData<S>? getState<S>(TinkActionLike action, String name);

  void next();
}
mixin ActionState {}
