import 'package:hypi_tink_api/hypi_tink_api.dart';
import 'package:ulid/ulid.dart';

export 'package:hypi_tink_api/variable.dart';

const String editorError = 'tink.editor.error';
const String editorAppInstanceChanged = 'tink.editor.app.instance.changed';

mixin CASMLike {
  CASMLike on<S>(
    String name,
    CASMEventHandler<S> f, {
    TinkStateEventType? type,
    bool emitLatest = true,
    bool subscribe = true,
    TinkLivenessHandler? state,
  });

  CASMLike emit<S>(TinkStateEvent<S> current);
}

mixin CASMEventLike<T> {
  String get eventName;

  bool get asEventProvider => false;

  TinkStateEvent<T> get asEvent => TinkStateEvent(
        eventName,
        this as T,
        TinkStateEventType.update,
        asEventProvider,
      );
}

mixin TinkEditorEvent<V> implements TinkVariableData<V> {
  ///If true, this event will be kept and when asked and a more recent value isn't available, it will be used as the current value
  @override
  bool asProvider = false;
  static const String editorAppTinkAppModel = 'tink.editor.app.model';
  static const String editorAppAccount = 'tink.editor.app.account';
  static const String editorCurrentScreen = 'tink.editor.screen.current';
  static const String editorCurrentComponent = 'tink.editor.component.current';
  static const String editorSelectedWidget = 'tink.editor.selected.widget';
  static const String editorRenderMode = 'tink.editor.render.mode';
  static const String editorRebuild = 'tink.editor.rebuild';
  static const String editorShowBottomSheet = 'tink.editor.bottomSheet';
  static const String editorBottomSheetOptions =
      'tink.editor.bottomSheet.options';
  static const String editorThemeModel = 'tink.editor.theme.model';
  static const String editorLogicConfig = 'tink.editor.logic.builder.config';
  static const String editorEvents = 'tink.editor.events';
// static String editorAppResolvedConfig = 'tink.editor.app.config.resolved';
// topRightActivityEvent,
// rightPaneView,
// footerShowLeftActions,
// footerShowRightActions,
// footerShowWorkflow,
}

typedef CASMEventHandler<S> = void Function(
  TinkStateEvent<S> prev,
  TinkStateEvent<S> current,
);

typedef CASMValueHandler<S> = TinkStateEvent<S> Function(
  TinkStateEvent<S> prev,
  TinkStateEvent<S> current,
);

typedef CASMValueProvider<S> = List<TinkStateEvent<S>> Function();

mixin TinkLivenessHandler {
  int get id;

  BoolCallback get isSubscriberAlive;
}

typedef BoolCallback = bool Function();

class TinkCASMStatus with TinkLivenessHandler {
  @override
  final BoolCallback isSubscriberAlive;
  @override
  final int id;

  TinkCASMStatus(this.isSubscriberAlive, Object source) : id = source.hashCode;

  factory TinkCASMStatus.alwaysAlive(Object source) =>
      TinkCASMStatus(() => true, source);

  factory TinkCASMStatus.neverAlive() => TinkCASMStatus(() => false, 0);
}

class CASMErrPair {
  Error? error;
  Exception? exception;
  StackTrace trace;

  CASMErrPair.except(this.exception, this.trace);

  CASMErrPair.error(this.error, this.trace);
}

abstract class TinkStateListener<S> {
  ///The type of events this listener wants to receive. If null, it gets all types
  TinkStateEventType? type;

  ///The name of specific events this listener wants to receive. If null, it gets all event names
  String? name;

  ///If true, this listener is interested in any event that starts with this name
  bool prefix;
  BoolCallback? alive;
  final Set<String> events = {};

  TinkStateListener(
      this.name, this.prefix, this.type, TinkLivenessHandler handler)
      : alive = (() => handler.isSubscriberAlive());

  bool wants<O>(TinkStateEvent<O> event) {
    return //
        (type == null || type == event.type) &&
            (name == null ||
                name == event.name ||
                (prefix && name != null && event.name.startsWith(name!))) &&
            event.value is S;
  }

  void unsubscribed() {}

  void accept(TinkStateEvent<S> prev, TinkStateEvent<S> current);
}

///Represents someone...or something performing operations in the app
class Actor {
  String id;
  String displayName;
  double x;
  double y;

  Actor({
    required this.id,
    required this.displayName,
    required this.x,
    required this.y,
  });
}

enum TinkStateEventType { update, delete, error, notification, undo, redo }

class TinkStateEvent<S> {
  Ulid id;
  String name;
  S value;
  Actor actor;
  TinkStateEventType type;

  ///If true, this event will be kept and when asked if a more recent value isn't available, it will be used as the current value
  bool asProvider;

  ///If set and returns false when [asProvider] is true then it causes this to be unregistered as a provider for its value
  BoolCallback? alive;
  static final Actor unknown = Actor(
    id: 'unknown',
    displayName: 'unknown',
    x: 0,
    y: 0,
  );

  TinkStateEvent._(
    this.id,
    this.name,
    this.value,
    this.actor,
    this.type,
    this.asProvider,
  );

  factory TinkStateEvent(
      String name, S value, TinkStateEventType type, bool asProvider,
      {Ulid? id, Actor? actor}) {
    return TinkStateEvent._(
      id ??= Ulid(),
      name,
      value,
      actor ??= unknown,
      type,
      asProvider,
    );
  }

  factory TinkStateEvent.error(S value, TinkStateEventType type,
      {Ulid? id, Actor? actor}) {
    return TinkStateEvent._(
        id ??= Ulid(), 'error', value, actor ??= unknown, type, false);
  }

  TinkStateEvent<S> asUndo() {
    return TinkStateEvent._(
      Ulid(),
      name,
      value,
      actor,
      TinkStateEventType.undo,
      false,
    );
  }

  TinkStateEvent<S> asRedo() {
    return TinkStateEvent._(
      Ulid(),
      name,
      value,
      actor,
      TinkStateEventType.redo,
      false,
    );
  }
}

class TinkEditorError {
  final Error e;
  final StackTrace trace;

  TinkEditorError(this.e, this.trace);
}
