import 'dart:async';
import 'dart:convert';

import 'package:flutter/material.dart';
import 'package:hypi_tink_api/hypi_tink_api.dart';

class HypiMeta {
  String id;
  String? instanceId;
  String? createdBy;
  late DateTime created = DateTime.now();
  DateTime? updated;

  HypiMeta({
    required this.id,
    this.instanceId,
    this.createdBy,
    this.updated,
  });

  Map<String, dynamic> get forServer =>
      toJSON..removeWhere((key, value) => !['id', 'instanceId'].contains(key));

  Map<String, dynamic> get toJSON => {
        'id': id,
        'instanceId': instanceId,
        'createdBy': createdBy,
        'created': created.toUtc().toIso8601String(),
        'updated': updated?.toUtc().toIso8601String(),
      };
}

const String tinkRefProtocol = 'tink-ref://';

String tinkRefToId(String ref) => ref.substring(11);

bool isTinkRef(String str) => str.startsWith(tinkRefProtocol);

String entityRef(TinkEntity e) => entityRefFromId(e.hypi.id);

String entityRefFromId(String id) => '$tinkRefProtocol$id';

String entitySaveEventName(String id) =>
    '$tinkEntityPrefix.save.${entityRefFromId(id)}';

String entityUpdateEventName(String id) =>
    '$tinkEntityPrefix.update.${entityRefFromId(id)}';

Map<String, dynamic> entityFrame(
    List<MapEntry<String, dynamic>> fields,
    HypiMeta hypi,
    String entityType,
    String entitySubType,
    String entityGroupId,
    String appInstanceId,
    {Map<String, dynamic>? target}) {
  target = target ?? {};
  target.addAll({
    'fields': fields
        .map((e) => {
              'name': e.key,
              //The value is a Hypi JSON type so must be written as such in all cases
              'value': json.encode(e.value),
              'hypi': {'id': 'fld.${e.key}.${hypi.id}'}
            })
        .toList(),
    'hypi': hypi.toJSON,
    'rootEntityID': appInstanceId,
    'entityType': entityType,
    'entitySubType': entitySubType,
    'entityGroupId': entityGroupId,
  });
  return target;
}

Map<String, Map<String, dynamic>> fieldsAsMap(Map<String, dynamic> obj) {
  return Map.fromEntries(((obj['fields'] as List?) ?? []).map((e) {
    final val = e as Map<String, dynamic>;
    return MapEntry(val['name'], val);
  }));
}

Map<String, dynamic> fieldsFromDB(resolved) {
  final fields = fieldsAsMap(resolved)
      //The value['value'] is because resolved returns the entity field from the DB
      .map((key, value) => MapEntry(key, value['value']));
  return fields;
}

typedef TinkFieldWriter = Function(TinkEntity entity, TinkEntitySet entities);
typedef TinkFieldReader = Future<T?> Function<T extends TinkEntity>(
  ///When called, should return a new instance of the entity with default or uninitalised values
  ///May not be called if the entity being requested has already been de-serialised.
  ///In such case, the existing entity will be returned by the resulting future
  T Function() creator,

  ///The value from which to inflate the object returned by [creator]
  dynamic,
);

mixin TinkEntityField {
  String get name;

  ///Must return a value serialisable by [json].
  ///i.e. one of number, boolean, string, null, list or a map with string keys
  dynamic write(TinkFieldWriter ser, TinkEntitySet entities);

  ///Deserialise this field into the given entity
  Future<dynamic> read(EntityEventSource src, dynamic value);

  ///If the value returned by [write] is not directly serialisable then this is called and it should return a serialisable value
  ///If the value returned still isn't then .toJSON is called on the value returned by [write]
  ///Object? Function(dynamic)? toEncodable;
}

typedef TinkEntitySet = Map<String, Map<String, dynamic>>;

mixin EntityEventSource {
  ///The unique ID of the entity this source
  String get entityId;

  Map<String, Map<String, dynamic>> get fieldsByName;

  ///The reader to use to deserialise entity fields
  TinkFieldReader get read;

  Future<dynamic> ensureResolved(dynamic input);

  ///To get the value of a field. For primitives i.e. String, int, double etc
  ///This returns immediately. For non-primitives, it may result in a network request to load the value of that field
  Future<dynamic> valueOf(String fieldName);

  ///Provides a mechanism for fetching all values of all fields in this entity.
  ///The future returned will complete when all fields have been fetched and successfully passed to [onField].
  ///Note that for large numbers of fields or data, several network requests may be made to produce all the data needed.
  ///The implementation may, at its discretion fetch the data via individual requests or over a persistent stream like web sockets.
  Future<void> allFields(
    Future<dynamic> Function(String fieldName, dynamic value) onField,
  );
}

const String tinkEntityPrefix = 'tink.entity.eventing';

HypiMeta hypiMeta(Map<String, dynamic> e) {
  return HypiMeta(
    id: e['id']!,
    instanceId: e['instanceId'],
    createdBy: e['createdBy'],
    updated: e['updated'] == null ? null : DateTime.parse(e['updated']!),
  )..created =
      e['created'] == null ? DateTime.now() : DateTime.parse(e['created']!);
}

class EntityEventSourceState extends TinkKV<EntityEventSource>
    with TinkVariableData<EntityEventSource> {
  EntityEventSourceState(EntityEventSource v)
      : super(entityUpdateEventName(v.entityId), v);
}

/*
class EntityPersistenceState extends TinkKV<TinkEntity>
    with TinkVariableData<TinkEntity> {
  EntityPersistenceState(TinkEntity v)
      : super(entitySaveEventName(v.hypi.id), v);
}
*/

///Any object that needs to be persisted and reloaded into the editor later
mixin TinkEntity {
  //static final _scheduler = Timer.periodic(const Duration(milliseconds: 100), (timer) {});

  HypiMeta get hypi;

  set hypi(HypiMeta val);

  ///type = Handler
  String get entityType;

  ///subtype = ScreenHandler
  String get entitySubType;

  ///groupId = Widget | Screen | Component
  String get entityGroupId;

  List<TinkEntityField> get fields;

  bool canPersist = true;

  //Duration entityAutoSaveDuration = const Duration(milliseconds: 250);
  //DateTime? entityLastSaved;

  late Map<String, TinkEntityField> fieldsByName =
      Map.fromEntries(fields.map((e) => MapEntry(e.name, e)));

  BoolCallback? serialisableIsAlive;

  bool _registered = false;

  //Called after the entity is deserialised - enables initialisation
  @mustCallSuper
  void postDeserialisation() {}

  save(CASMLike casm) {
    subscribeToUpdates(casm);
    //emit a save event for this entity
    casm.emit(
      TinkStateEvent(
        entitySaveEventName(hypi.id),
        this,
        TinkStateEventType.update,
        false, //don't act as a provider, this is one and done
      ),
    );
  }

  subscribeToUpdates(CASMLike casm) {
    //only subscribe once otherwise will end up being called multiple times for a single update
    if (!_registered) {
      casm.on<EntityEventSourceState>(entityUpdateEventName(hypi.id),
          (prev, current) {
        onUpdate(current.value.value);
      }, state: TinkCASMStatus(serialisableIsAlive ?? () => false, this));
      _registered = true;
    }
  }

  ///When invoked, the entity should use the event source to obtain the most recent value for each of its fields, internally performing an update where necessary
  ///Implementations can override this if for example they only want to update selected fields they can use [EntityEventSource.valueOf]
  void onUpdate(EntityEventSource src) {
    //Do not allow objects that are partially loaded to be persisted
    canPersist = false;
    src.allFields((fieldName, value) {
      Future<dynamic> fieldFuture;
      final field = fieldsByName[fieldName];
      if (field != null && value != null) {
        fieldFuture = field.read(src, value);
      } else {
        fieldFuture = Future.value(null);
      }
      return fieldFuture;
    }).whenComplete(() {
      postDeserialisation();
      //Done de-serialising, this can now be persisted
      canPersist = true;
    });
  }

  @override
  bool operator ==(Object other) =>
      identical(this, other) ||
      other is TinkEntity &&
          runtimeType == other.runtimeType &&
          hypi.id == other.hypi.id;

  @override
  int get hashCode => hypi.id.hashCode;
}

class MapEntityField<T extends TinkEntity> with TinkEntityField {
  @override
  String name;

  @override
  Future<dynamic> read(EntityEventSource src, value) {
    if (value == null) {
      return Future.value(null);
    } else {
      if (value is Map) {
        return Future.wait(value.entries.map((pair) {
          return src.ensureResolved(pair.value).then((resolved) {
            return src.read(() => creator(resolved), resolved).then((read) {
              if (read != null) onEntry(pair.key, read);
            });
          }).catchError((e) => throw e);
        }));
      } else {
        throw UnsupportedError(
          'Attempted to de-serialise map from non-map value for field $name',
        );
      }
    }
  }

  @override
  write(TinkFieldWriter ser, TinkEntitySet entities) => doWrite(ser, entities);

  ///Use fieldsFromDB(resolved) to get a map of the expected key value pairs
  T Function(dynamic) creator;
  Function(String key, T val) onEntry;
  dynamic Function(
    Function(TinkEntity, TinkEntitySet entities),
    TinkEntitySet entities,
  ) doWrite;

  MapEntityField({
    required this.name,
    required this.creator,
    required this.onEntry,
    required this.doWrite,
  });
}

class ListEntityField<T extends TinkEntity> with TinkEntityField {
  @override
  String name;

  @override
  Future<dynamic> read(EntityEventSource src, value) {
    if (value == null) {
      return Future.value(null);
    } else {
      if (value is List) {
        return Future.wait(value.map((val) {
          return src.ensureResolved(val).then((resolved) {
            return src.read(() => creator(resolved), resolved).then((read) {
              if (read != null) onEntry(read);
            });
          }).catchError((e) => throw e);
        }));
      } else {
        throw UnsupportedError(
          'Attempted to de-serialise list from non-list value for field $name',
        );
      }
    }
  }

  @override
  write(TinkFieldWriter ser, TinkEntitySet entities) => doWrite(ser, entities);

  ///Use fieldsFromDB(resolved) to get a map of the expected key value pairs
  T Function(dynamic) creator;
  Function(T val) onEntry;
  dynamic Function(
    Function(TinkEntity, TinkEntitySet entities),
    TinkEntitySet entities,
  ) doWrite;

  ListEntityField({
    required this.name,
    required this.creator,
    required this.onEntry,
    required this.doWrite,
  });
}

class EntityField with TinkEntityField {
  @override
  String name;

  @override
  Future<dynamic> read(EntityEventSource src, value) {
    if (value == null) {
      return Future.value(null);
    } else {
      return src.ensureResolved(value).then((resolved) {
        return doRead(src, resolved);
      }).catchError((e) => throw e);
    }
  }

  @override
  write(TinkFieldWriter ser, TinkEntitySet entities) => doWrite(ser, entities);

  ///Use fieldsFromDB(resolved) to get a map of the expected key value pairs
  Future<dynamic> Function(
    EntityEventSource src,
    dynamic raw,
    //Map<String, dynamic> resolved,
  ) doRead;
  dynamic Function(
    Function(TinkEntity, TinkEntitySet entities),
    TinkEntitySet entities,
  ) doWrite;

  EntityField({
    required this.name,
    required this.doRead,
    required this.doWrite,
  });
}

class TSimpleField with TinkEntityField {
  @override
  String name;

  @override
  Future<dynamic> read(EntityEventSource src, value) async {
    if (value == null) {
      return Future.value(null);
    } else {
      setter(value);
      return value;
    }
  }

  @override
  write(TinkFieldWriter ser, TinkEntitySet entities) => getter();
  void Function(dynamic) setter;
  dynamic Function() getter;

  TSimpleField(this.name, this.setter, this.getter);
}

class TDoubleField extends TSimpleField {
  TDoubleField(
    String name,
    void Function(dynamic) setter,
    dynamic Function() getter,
  ) : super(name, (value) => setter(value is int ? value.toDouble() : value),
            getter);
}
