import 'package:flutter/widgets.dart';
import 'package:hypi_tink_api/casm.dart';
import 'package:hypi_tink_api/hypi_tink_api.dart';
import 'package:hypi_tink_api/models.dart';
import 'package:hypi_tink_api/target_device.dart';

///Many widgets are stateless but Hypi provides the ability to update their attributes
///When that happens, we must rebuild the screen for the changed attributes to show up
///If your widget handler is stateful just call setState locally, no need for a global rebuild unless it affects other parts of the UI
///It is in effect the same as setState in a stateful widget but does not incur any overhead unless the widget's attribute actually changes
class TinkRebuildScreen implements TinkRuntimeEvent<TinkRebuildScreen> {
  static final String eventName = eventPrefix + '.screen.rebuild';
  @override
  bool asProvider = false;

  @override
  String key = eventName;

  @override
  late final value = this;
  final Function() fn;

  TinkRebuildScreen(this.fn);

  late TinkStateEvent<TinkRebuildScreen> asEvent =
      TinkStateEvent(key, value, TinkStateEventType.update, asProvider);
}

mixin TinkAttributeEvent implements TinkRuntimeEvent<dynamic> {
  TinkWidgetHandler get handler;

  TinkWidgetAttribute get attr;

  dynamic get data;

  @override
  bool asProvider = false;
}

class TinkWidgetRuntimeContext {
  BuildContext context;
  void Function(Function()) setState;

  TinkWidgetRuntimeContext(this.context, this.setState);
}

typedef ChildWidgetValue = Either<TinkWidgetHandler, List<TinkWidgetHandler>>;

///Some widgets have a single child and hence need only the property name to that child
///whilst others have a list of children and need a name to that list hence either acts as a unifying type to achieve both
class ChildWidget {
  ///Name of the property/argument that the associated child/children are placed under
  String name;
  ChildWidgetValue value;

  ChildWidget(this.name, this.value);
}

mixin TreeNodeHandler {
  onNodeTreeAccept(TreeNodeHandler handler);

  onNodeTreeWillAccept(TreeNodeHandler? dropped);

  onNodeTreeMove(DragTargetDetails<TreeNodeHandler> details);

  onNodeTreeLeave(TreeNodeHandler? details);

  onNodeTreeAcceptWithDetails(DragTargetDetails<TreeNodeHandler> details);
}

mixin TinkWidgetHandler
    implements TinkEntity, CASMLike, Displayable, TreeNodeHandler {
  ///An implementation that is guaranteed to provide a default for all APIs in this mixin
  late TinkWidgetHandler delegate;

  ///The handler that, if this is the delegate, this delegates for. If it is not the delegate self == this
  late TinkWidgetHandler self;

  ///Called to provide an initialisation point for delegate implementations.
  ///Must be called by implementations or they must initialise delegate and self themselves AND call this on the delegate
  void initDelegate(TinkWidgetHandler delegate, TinkWidgetHandler self) {
    this.delegate = delegate;
    this.self = self;
    delegate.initDelegate(delegate, self);
  }

  @override
  String get groupId => delegate.groupId;

  @override
  String get label => delegate.label;

  WidgetDefModel get def => delegate.def;

  set def(WidgetDefModel def) => delegate.def = def;

  TinkWidgetHandler get parent => delegate.parent;

  set parent(TinkWidgetHandler parent) => delegate.parent = parent;

  Set<EventModel> get events => delegate.events;

  set events(Set<EventModel> events) => delegate.events = events;

  Map<String, dynamic> get args => delegate.args;

  set args(Map<String, dynamic> _args) => delegate.args = _args;

  GlobalKey get key => delegate.key;

  Set<TinkWidgetAttribute> get attributes => delegate.attributes;

  @override
  String get id => delegate.id;

  set id(String id) => delegate.id = id;

  @override
  bool get canPersist => delegate.canPersist;

  @override
  set canPersist(bool _canPersist) => delegate.canPersist = _canPersist;

  TargetDevice get device => delegate.device;

  set device(TargetDevice _device) => delegate.device = _device;

  bool get settingHandle => delegate.settingHandle;

  set settingHandle(bool _settingHandle) =>
      delegate.settingHandle = _settingHandle;

  bool get highlight => delegate.highlight;

  set highlight(bool _highlight) => delegate.highlight = _highlight;

  bool? get isChildFieldAList => delegate.isChildFieldAList;

  set isChildFieldAList(bool? _isChildFieldAList) =>
      delegate.isChildFieldAList = _isChildFieldAList;

  String? get addChildFieldName => delegate.addChildFieldName;

  ScreenModel? get currentScreen => delegate.currentScreen;

  set addChildFieldName(String? addChildFieldName) =>
      delegate.addChildFieldName = addChildFieldName;

  @override
  Map<String, TinkEntityField> get fieldsByName => delegate.fieldsByName;

  @override
  set fieldsByName(Map<String, TinkEntityField> _fieldsByName) =>
      delegate.fieldsByName = _fieldsByName;

  @override
  HypiMeta get hypi => delegate.hypi;

  @override
  set hypi(HypiMeta val) => delegate.hypi = val;

  @override
  BoolCallback? get serialisableIsAlive => delegate.serialisableIsAlive;

  bool _listeningForAttrChanges = false;
  TinkWidgetRuntimeContext? currentRuntimeCtx;

  @override
  set serialisableIsAlive(BoolCallback? _serialisableIsAlive) =>
      delegate.serialisableIsAlive = _serialisableIsAlive;

  @override
  String get entityGroupId => delegate.entityGroupId;

  @override
  String get entitySubType => delegate.entitySubType;

  @override
  String get entityType => delegate.entityType;

  @override
  List<TinkEntityField> get fields => delegate.fields;

  List<ChildWidget> get children => delegate.children;

  bool get hasChildren => children.isNotEmpty;

  TinkRenderMode get renderMode => delegate.renderMode;

  bool get isRuntimeMode =>
      renderMode == TinkRenderMode.preview ||
      renderMode == TinkRenderMode.runtime;

  void setArg(String name, dynamic value) => delegate.setArg(name, value);

  void emitEvent(EventModel event) => delegate.emitEvent(event);

  void triggerEditorRebuild(VoidCallback fn) =>
      delegate.triggerEditorRebuild(fn);

  void listenForAttributeChanges() {
    if (!_listeningForAttrChanges) {
      for (final attr in attributes) {
        on<TinkAttributeEvent>(attrName(this, attr.id.name), (_, val) {
          setAttr(attr, val.value);
        });
      }
      _listeningForAttrChanges = true;
    }
  }

  ///Called when the given attribute has been given a new value.
  ///This assumes the attribute can/should be directly set in the args passed to the widget
  void setAttr(TinkWidgetAttribute attr, TinkAttributeEvent current) {
    rerender(() => args[attr.id.name] = current.data);
  }

  ///If this widget has setup a runtime context then call setState which will trigger a rebuild
  ///If it hasn't then the whole screen will be re-built to force the update to become visible to the user
  ///Obviously rebuilding the screen is more expensive so most widgets should provide a [TinkWidgetRuntimeContext] when in runtime mode.
  void rerender(Function() fn) {
    if (currentRuntimeCtx == null) {
      emit(TinkRebuildScreen(fn).asEvent);
    } else {
      currentRuntimeCtx?.setState(fn);
    }
  }

  Widget render() => delegate.render();

  Widget showWidgetProperties(BuildContext context, StateSetter setState) =>
      delegate.showWidgetProperties(context, setState);

  ///Widgets that can have other widgets dropped on them should set this to true
  ///If it is false, this handler's widget will not be considered for drag/drop events
  bool canAcceptChildren = false;

  TinkWidgetHandler? intersects(Offset activityPosition) =>
      delegate.intersects(activityPosition);

  ///Called when a new instance must be created to render on screen
  @mustCallSuper
  void droppedOnRootCanvas<T extends WidgetDefModel>(
    T def,
    Offset offset,
    dynamic data,
  ) =>
      delegate.droppedOnRootCanvas(def, offset, data);

  bool onMove(Offset offset, dynamic data) => delegate.onMove(offset, data);

  bool onLeave() => delegate.onLeave();

  ///[details] is either [DragTargetDetails<dynamic>] or a string that is the name of the widget dropped or if it was re-arranged then it is the widget itself that was dropped
  bool canHandleDropped(Offset offset, dynamic data) =>
      delegate.canHandleDropped(offset, data);

  void handleDropped(Offset offset, dynamic data) =>
      delegate.handleDropped(offset, data);

  @override
  CASMLike emit<S>(TinkStateEvent<S> current) => delegate.emit(current);

  @override
  CASMLike on<S>(
    String name,
    CASMEventHandler<S> f, {
    TinkStateEventType? type = TinkStateEventType.update,
    bool emitLatest = true,
    bool subscribe = true,
    TinkLivenessHandler? state,
  }) =>
      delegate.on(
        name,
        f,
        type: type,
        emitLatest: emitLatest,
        subscribe: subscribe,
        state: state,
      );

  @override
  void onUpdate(EntityEventSource src) => delegate.onUpdate(src);

  @override
  void postDeserialisation() => delegate.postDeserialisation();

  @override
  save(CASMLike casm) => delegate.save(casm);

  @override
  subscribeToUpdates(CASMLike casm) => delegate.subscribeToUpdates(casm);

  void canAcceptChildNodes({
    required String childFieldName,
    required bool isList,
  }) =>
      delegate.canAcceptChildNodes(
        childFieldName: childFieldName,
        isList: isList,
      );

  void setNonListChild(String argName, TinkWidgetHandler child) =>
      delegate.setNonListChild(argName, child);

  void setChild(String name, ChildWidgetValue value) =>
      delegate.setChild(name, value);

  TinkWidgetHandler constructTree(
          String widgetName,
          Map<String, Either<TinkWidgetHandler, List<TinkWidgetHandler>>>
              children,
          {Map<String, dynamic>? props}) =>
      delegate.constructTree(widgetName, children, props: props);

  EntityField childEntityField(String fieldName) =>
      delegate.childEntityField(fieldName);

  void removeChildByName(String name) => delegate.removeChildByName(name);

  void swapChild(TinkWidgetHandler oldChild, TinkWidgetHandler newChild) =>
      delegate.swapChild(oldChild, newChild);

  void removeChild(TinkWidgetHandler oldChild) =>
      delegate.removeChild(oldChild);

  ChildWidgetValue setChildIfAbsent(
    String name,
    ChildWidgetValue Function() value,
  ) =>
      delegate.setChildIfAbsent(name, value);

  void addChildHandler(TinkWidgetHandler child) =>
      delegate.addChildHandler(child);

  ChildWidgetValue? getChildByName(String name) =>
      delegate.getChildByName(name);

  Widget highlightWidget(Widget widget) => delegate.highlightWidget(widget);

  void showSettingsHandle() => delegate.showSettingsHandle();

  @override
  onNodeTreeAccept(TreeNodeHandler handler) =>
      delegate.onNodeTreeAccept(handler);

  @override
  onNodeTreeAcceptWithDetails(DragTargetDetails<TreeNodeHandler> details) =>
      delegate.onNodeTreeAcceptWithDetails(details);

  @override
  onNodeTreeLeave(TreeNodeHandler? details) =>
      delegate.onNodeTreeLeave(details);

  @override
  onNodeTreeMove(DragTargetDetails<TreeNodeHandler> details) =>
      delegate.onNodeTreeMove(details);

  @override
  onNodeTreeWillAccept(TreeNodeHandler? dropped) =>
      delegate.onNodeTreeWillAccept(dropped);
}
