import 'package:hypi_tink_api/hypi_tink_api.dart';

enum TinkVariableType {
  text,
  bool,
  int,
  double,
  num,
  json,
  variable,

  ///Any of the types in this list, NOT any dart type
  any
}

extension TinkVariableTypeExten on TinkVariableType {
  static TinkVariableType? of(String name) {
    switch (name.toLowerCase().trim()) {
      case 'text':
        return TinkVariableType.text;
      case 'bool':
        return TinkVariableType.bool;
      case 'int':
        return TinkVariableType.int;
      case 'double':
        return TinkVariableType.double;
      case 'num':
        return TinkVariableType.num;
      case 'json':
        return TinkVariableType.json;
      default:
        return null;
    }
  }

  String name() {
    switch (this) {
      case TinkVariableType.text:
        return 'text';
      case TinkVariableType.bool:
        return 'bool';
      case TinkVariableType.int:
        return 'int';
      case TinkVariableType.double:
        return 'double';
      case TinkVariableType.num:
        return 'num';
      case TinkVariableType.json:
        return 'json';
      case TinkVariableType.any:
        return 'any';
      case TinkVariableType.variable:
        return 'variable';
    }
  }
}

///Any data that can be emitted by the application can be stored as variable.
///Application variable can be bound in the editor at development time and at run time the values are fetched and used as inputs to workflows
///Note that some [TinkEditorHandler]s like those for form elements have default workflows which bind their values to some variable
mixin TinkVariableData<V> {
  ///If true, this event will be kept and when asked and a more recent value isn't available, it will be used as the current value
  bool asProvider = false;

  String get key;

  V get value;
}

abstract class TinkKV<V> with TinkVariableData<V> {
  final String _k;
  final V _v;

  TinkKV(this._k, this._v);

  @override
  String get key => _k;

  @override
  V get value => _v;
}

class TinkIntVariable extends TinkKV<int> {
  TinkIntVariable(String name, int value) : super(name, value);
}

class TinkVariableRef extends TinkKV<VariableRef> {
  TinkVariableRef(String name, VariableRef value) : super(name, value);
}

class TinkTextVariable extends TinkKV<String> {
  TinkTextVariable(String name, String value) : super(name, value);
}

class TinkBoolVariable extends TinkKV<bool> {
  TinkBoolVariable(String name, bool value) : super(name, value);
}

class TinkDoubleVariable extends TinkKV<double> {
  TinkDoubleVariable(String name, double value) : super(name, value);
}

class TinkNumVariable extends TinkKV<num> {
  TinkNumVariable(String name, num value) : super(name, value);
}

class TinkJsonVariable<T> extends TinkKV<dynamic> {
  TinkJsonVariable(String name, dynamic value) : super(name, value);
}

mixin WorkflowModel implements Displayable {
  ///Set of actions in this workflow
  final List<TinkActionLike> actions = [];

  ///Set of environments created for this workflow. Each action WITHIN the workflow has its own
  final Map<TinkActionLike, ActionEnv> envs = {};
}

///Any type which can be in an invalid state - this provides the interface to knowing if it is valid or not
mixin Validatable {
  bool get valid;
}

///Marker interface to identify all bindings
mixin Binding implements Validatable, TinkEntity {
  String get id;

  String? label;

  VariableRef get ref;

  @override
  bool operator ==(Object other) {
    return other is Binding ? id == other.id : false;
  }

  @override
  int get hashCode => id.hashCode;
}

///The location where the variable is held. This determines where it can be accessed and in the case of [VariableScope.device] for how long
enum VariableScope {
  ///The variable is stored locally on the device's persistent media e.g SD card, disk, localstorage.
  ///When it is read, it is read from memory if cached otherwise from local persistent media
  ///When written, it is always written to the local persistent media before subscribers are notfied
  ///The exact mechanism by which it is stored varies depending on the target device.
  ///iOS - NSUserDefaults
  ///macOS - NSUserDefaults
  ///Android - SharedPreferences
  ///Web - LocalStorage
  ///Linux - A file
  ///Windows - A file
  device,

  ///State with this scope is available to all screens. Updating it on one screen will update it on all of them. In memory only.
  app,

  ///State with this scope is available to all screens. Updating it on one screen will update it on all of them. In memory only.
  static,

  ///State with this scope is available only on the screen where it is declared, therefore two screens that have a variable with the same name
  ///are independent and updating it one one screen does not update it on the other. In memory only.
  screen,

  //State with this scope is available only within a workflow. It allows actions to create variable that are not automatically
  workflow,
}

extension VariableScopeExt on VariableScope {
  String name() {
    switch (this) {
      case VariableScope.device:
        return 'device';
      case VariableScope.app:
        return 'app';
      case VariableScope.screen:
        return 'screen';
      case VariableScope.workflow:
        return 'workflow';
      case VariableScope.static:
        return 'static';
    }
  }

  static VariableScope? of(String str) {
    switch (str) {
      case 'device':
        return VariableScope.device;
      case 'app':
        return VariableScope.app;
      case 'screen':
        return VariableScope.screen;
    }
  }
}

///A function that is called when a source produces a value that should be pushed to the variables that are bound to it
typedef VariableSourceListener = void Function(dynamic value);

///The source of a variable is anything that can produce it.
///A widget for example can produce the value of one of its attributes which is assigned to a variable
///A workflow action as well can output values that is bound as the source of a variable
mixin VariableSource implements Binding, Validatable {
  ///A source can be enabled/disabled. When false the source will not produce updates
  bool get enabled;

  dynamic get value;

  ///For the given variable, produce a string description of yourself
  String desc(TinkVariable variable);

  String get eventKey;
}

///A variable target is anything that wants to be told when that variable has changed.
///An attribute of a widget for example can be bound to a target
///This is for "push" based targets. Note that workflow actions use a "pull" based approach where they request the latest value of a variable at any point during their evaluation
mixin VariableTarget implements Binding {
  ///A target can be enabled/disabled. When false the target will not receive updates
  bool get enabled;
}

///A variable is a named location where information can be stored.
///The information stored can be used to update parts of the UI or sent in API requests
mixin TinkVariable implements Displayable {
  ///The name of the variable
  String get name;

  ///The type of the variable. All variables must have a type matching one of these.
  TinkVariableType get type;

  ///The scope of a variable determines where it is accessible
  VariableScope get scope;

  ///A differentiator helps to specify the uniqueness of the variable.
  ///In the case of a [VariableScope.screen] this is the path of the screen
  ///e.g. .screen-id would produce an ID like
  ///tink.variable.screen-id.screen.x where the screen is the scope and x is the name.
  ///For screen scope => .screen-id
  ///For device scope => device
  ///For app    scope => app
  ///Only screen scope changes.
  String get differentiator;

  ///The sources that can update this variable
  List<VariableSource> get sources;

  ///The set of targets that will be told when this variable's value changes
  List<VariableTarget> get targets;

  dynamic get latest;

  TinkScreenId? get screenId;

  ///Can be used to uniquely identify this variable in any context
  VariableRef get ref;

  @override
  bool operator ==(Object other) {
    return other is TinkVariable ? id == other.id : false;
  }

  @override
  int get hashCode => id.hashCode;
}

class VariableRef with TinkEntity {
  String name;
  VariableScope scope;
  TinkVariableType type;
  String? screenId;
  String differentiator;
  dynamic defaultValue;

  String get id =>
      'tink.variable.${scope.name()}.${type.name()}.$differentiator.$name';

  @override
  late String entityType = 'VariableRef';
  @override
  late String entitySubType = scope.name();
  @override
  late String entityGroupId = 'state';

  VariableRef({
    required this.name,
    required this.scope,
    required this.type,
    required this.differentiator,
    this.screenId,
  });

  @override
  String toString() => id;

  @override
  bool operator ==(Object other) =>
      identical(this, other) ||
      other is VariableRef &&
          runtimeType == other.runtimeType &&
          id == other.id;

  @override
  int get hashCode => id.hashCode;

  @override
  late HypiMeta hypi = HypiMeta(id: id);

  @override
  late List<TinkEntityField> fields = [
    TSimpleField('name', (value) => name = value, () => name),
    TSimpleField('scope', (value) => scope = VariableScopeExt.of(value)!,
        () => scope.name()),
    TSimpleField('type', (value) => type = TinkVariableTypeExten.of(value)!,
        () => type.name()),
    TSimpleField('screenId', (value) => screenId = value, () => screenId),
    TSimpleField('differentiator', (value) => differentiator = value,
        () => differentiator),
  ];

/* @override
  void onUpdate(EntityEventSource src) {
    src.allFields((fieldName, value) {
      switch (fieldName) {
        case 'name':
          name = value;
          break;
        case 'scope':
          scope = VariableScopeExt.of(value) ?? scope;
          break;
        case 'type':
          type = TinkVariableTypeExten.of(value) ?? type;
          break;
        case 'differentiator':
          differentiator = value;
          break;
        case 'screenId':
          screenId = value;
          break;
      }
    });
  }*/
}

mixin TinkRuntimeEvent<V> implements TinkVariableData<V> {
  ///If true, this event will be kept and when asked and a more recent value isn't available, it will be used as the current value
  @override
  bool asProvider = false;
  static const String appState = 'tink.runtime.app.state';
}

///When a variable changes, emit this using the ref and the new value. Subscribers will be notified.
class VariableChangeEvent with TinkRuntimeEvent<VariableChangeEvent> {
  VariableRef ref;
  TinkVariableData data;

  VariableChangeEvent({
    required this.ref,
    required this.data,
  });

  @override
  String get key => ref.toString();

  @override
  VariableChangeEvent get value => this;
}

mixin TinkWorkflowBinding<E extends EventModel, W extends WorkflowModel,
    WM extends TinkWidgetHandler> implements Validatable, TinkEntity {
  ///Can be set to a function which indicates whether the binding is valid
  ///For example, in an action with a custom config it would return false until the custom settings are valid
  BoolCallback? isValidProvider;

  WM? get eventSrc;

  ///When this event is fired, evaluate the given workflow
  E? get event;

  ///The workflow to evaluate
  W get workflow;

  String get id;

  ///A binding of each input in the workflow. When the associated event is fired,
  ///these variables are retrieved and used as the input to start the workflow.
  Map<TinkActionLike, Map<String, VariableRef>> get bindings;
}

///Binds the value of an attribute to some variable
mixin TinkWidgetInputBinding<S extends ScreenModel,
    W extends TinkWidgetHandler> implements VariableTarget {
  ///The widget from which the attribute's value is taken when [event] occurs
  W? get attributeSrc;

  ///The attribute to set to the value of the variable
  TinkWidgetAttribute? get attribute;
}

const String eventPrefix = 'tink.event';
const String widgetEventPrefix = eventPrefix + '.widget';

String eventName(
  TinkWidgetHandler handler,
  EventModel event,
) =>
    widgetEventPrefix + '.' + handler.id + '.' + event.name;

String attrName(
  TinkWidgetHandler handler,
  String attrName,
) =>
    widgetEventPrefix + '.' + handler.id + '.' + attrName;

///Sets some variable when the given attribute is modified as a result of the given event
mixin TinkWidgetOutputBinding<E extends EventModel,
    W extends TinkWidgetHandler> implements VariableSource {
  ///The widget on which the [event] occurs
  W? get eventSrc;

  ///When this event is fired, bind the given attribute to the given variable
  E? get event;

  ///The widget from which the attribute's value is taken when [event] occurs
  W? get attributeSrc;

  ///The attribute to bind the given variable to
  TinkWidgetAttribute? get attribute;
}

///An attribute is a value that can be bound to a [TinkVariableData]
mixin TinkWidgetAttribute {
  ///The name and type of the attribute
  TinkParam get id;

  ///Given a value, produce a [TinkVariableData]
  dynamic get value;
}

class FNAttr with TinkWidgetAttribute {
  @override
  TinkParam id;
  dynamic Function() getter;

  FNAttr(this.id, this.getter);

  @override
  dynamic get value => getter();
}

typedef TinkScreenId = String;
